#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import math
import numpy as np

from sciveo.ml.images.tools import *


"""
Object Detection Bounding Boxes (bbox) of type [x, y, w, h]

If need to use [x1, y1, x2, y2] need to use the inverted convertor bbox_convert_inverted().

IoU between 2 object detections: iou(bbox1, bbox2)

"""

# convert from [x, y, w, h] -> [x1, y1, x2, y2]
def bbox_convert(bbox):
  x1 = bbox[0]
  y1 = bbox[1]
  x2 = x1 + bbox[2]
  y2 = y1 + bbox[3]
  return [x1, y1, x2, y2]

# convert from [x1, y1, x2, y2] -> [x, y, w, h]
def bbox_convert_inverted(bbox):
  x = bbox[0]
  y = bbox[1]
  w = bbox[2] - x
  h = bbox[3] - y
  return [x, y, w, h]

def bbox_norm(bbox, w, h):
  return (bbox[0] / w, bbox[1] / h, bbox[2] / w, bbox[3] / h)

def bbox_denorm(bbox, w, h):
  return (int(bbox[0] * w), int(bbox[1] * h), int(bbox[2] * w), int(bbox[3] * h))

def bbox_center(bbox):
  return (int(bbox[0] + bbox[2] / 2), int(bbox[1] + bbox[3] / 2))

def bbox_area(bbox):
  return bbox[2] * bbox[3]

def iou(bbox1, bbox2):
  x1 = max(bbox1[0], bbox2[0])
  y1 = max(bbox1[1], bbox2[1])
  x2 = min(bbox1[0] + bbox1[2], bbox2[0] + bbox2[2])
  y2 = min(bbox1[1] + bbox1[3], bbox2[1] + bbox2[3])

  if x1 < x2 and y1 < y2:
    a = (x2 - x1) * (y2 - y1)
  else:
    a = 0

  a1 = bbox_area(bbox1)
  a2 = bbox_area(bbox2)
  return a / (a1 + a2 - a)

def bbox_distance(bbox1, bbox2):
  return points_distance(bbox_center(bbox1), bbox_center(bbox2))


"""

Simple Draw object detectios helpers

"""
def image_shape(image):
  return image.shape[1], image.shape[0]

# Draw label bounding boxes of type [x, y, w, h], if [x1, y1, x2, y2] then set convert=False
def draw_label_bboxes(image, bboxes, color, convert=True):
  w, h = image_shape(image)
  for bbox in bboxes:
    if convert:
      bbox = bbox_convert(bbox)
    bbox = bbox_denorm(bbox, w, h)
    cv2.rectangle(image, (bbox[0], bbox[1]), (bbox[2], bbox[3]), color, 2, 1)
  return image



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import cv2
import gc
from PIL import Image

import torch
from transformers import AutoProcessor, AutoModelForCausalLM

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.tpu_base import *
from sciveo.media.pipelines.processors.image.generators import ImageToText
from sciveo.media.pipelines.base import ApiContent


class VideoToText(ImageToText):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)

  def process(self, media):
    debug("process", media['guid'])
    local_path = media["local_path"]

    cap = cv2.VideoCapture(local_path)
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    cap.set(cv2.CAP_PROP_POS_FRAMES, total_frames // 2)
    ret, frame = cap.read()
    cap.release()
    if ret:
      predict = self.predict_image_text(frame)
      return self.set_media(media, predict)
    else:
      return media

  def content_type(self):
    return "video"

  def name(self):
    return "video-to-text"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import boto3
import cv2
from PIL import Image

import torch
from torchvision import models, transforms

from sciveo.tools.logger import *
from sciveo.tools.common import *
from sciveo.media.pipelines.processors.tpu_base import *
from sciveo.media.pipelines.base import ApiContent


class ImageEmbedding(TPUBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)

    self.default.update({
      "model_id": 1,
      "output": False
    })

    cache_dir = os.path.join(os.environ['MEDIA_MODELS_BASE_PATH'], "models/")
    self.device = os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu")

    model_name = [
      "softel-resnet18-embedding.pth",
      "softel-resnet34-embedding.pth",
    ][self['model_id']]

    self.model_path = os.path.join(cache_dir, model_name)
    if os.path.isfile(self.model_path):
      debug(model_name, "available", self.model_path)
    else:
      debug("DWN", model_name)
      s3 = boto3.client('s3')
      s3.download_file("sciveo-model", model_name, self.model_path)

    self.api = ApiContent()

    self.preprocessor = transforms.Compose([
      transforms.Resize(256),
      transforms.CenterCrop(224),
      transforms.ToTensor(),
      transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
    ])

    self.model = torch.load(self.model_path)
    self.model.eval()

  def embed(self, local_path):
    image = Image.open(local_path)
    image_tensor = self.preprocessor(image).unsqueeze(0)
    with torch.no_grad():
      embedding = self.model(image_tensor)
    return embedding.squeeze().numpy()

  def process(self, media):
    debug("process", media['guid'])
    embedding = self.embed(media["local_path"])
    self.api.update(media, {"embedding_resnet_512": list(embedding)})
    return media

  def content_type(self):
    return "image"

  def name(self):
    return "image-embedding"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import time
import math
import numpy as np

from sciveo.tools.logger import *
from sciveo.ml.images.object_detection import *


"""

Object Detection Evaluation

"""
class EvalObjectDetection:
  def __init__(self, predictions, labels, class_names):
    self.predictions = predictions
    self.labels = labels
    self.class_names = class_names

  # Convert from yolo predictions
  def from_yolo(self):
    self.converted_predictions = []
    for predicted in predictions:
      boxes = predicted.boxes
      class_names = predicted.names

      converted_prediction = {}
      for i in range(len(boxes)):
        class_id = int(boxes.cls[i].item())
        label_class = class_names[class_id]
        if label_class not in self.class_names:
          continue
        box = boxes.xyxyn[i].tolist()
        confidence = boxes.conf[i].item()
        box_confidence = box + [confidence]

        converted_prediction.setdefault(label_class, [])
        converted_prediction[label_class].append(box_confidence)
      self.converted_predictions.append(converted_prediction)

    # TODO: currently labels of type [ {"class 1": [ ['x1':x1,'y1':y1,'x2':x2,'y2':y2] ... ], "class 2": [ ['x1':x1,'y1':y1,'x2':x2,'y2':y2] ... ]} ... ]
    self.converted_labels = []
    for label in self.labels:
      converted_label = {}
      for class_name, class_boxes in label.items():
        converted_label[class_name] = []
        for box in class_boxes:
          converted_label[class_name].append([box['x1'], box['y1'], box['x2'], box['y2']])
      self.converted_labels.append(converted_label)

  def compute_iou(self, box1, box2):
    """
    box1, box2: [x1, y1, x2, y2] format.
    Returns the IoU between the two boxes.
    """
    x1_inter = max(box1[0], box2[0])
    y1_inter = max(box1[1], box2[1])
    x2_inter = min(box1[2], box2[2])
    y2_inter = min(box1[3], box2[3])

    inter_area = max(0, x2_inter - x1_inter) * max(0, y2_inter - y1_inter)
    box1_area = (box1[2] - box1[0]) * (box1[3] - box1[1])
    box2_area = (box2[2] - box2[0]) * (box2[3] - box2[1])
    union_area = box1_area + box2_area - inter_area

    if union_area == 0:
      return 0.0

    return inter_area / union_area

  # Compute the Average Precision (AP)
  def compute_ap(self, class_name, iou_threshold=0.0):
    """
    Calculate the Average Precision based on IoU and confidence scores.
    Returns the AP for the current class.
    """

    count_labels = 0
    true_positives = []
    false_positives = []
    false_negatives = []
    detected = []

    for i, prediction in enumerate(self.converted_predictions):
      current_predictions = prediction.get(class_name, [])
      current_labels = self.converted_labels[i].get(class_name, [])

      count_labels += len(current_labels)
      TP = 0

      for prediction_box in current_predictions:
        max_iou = 0
        gt_match = None

        for label_box in current_labels:
          iou = self.compute_iou(prediction_box[:4], label_box)
          if iou > 0.5 and iou >= max_iou:
            max_iou = iou
            gt_match = label_box

        if max_iou >= iou_threshold and gt_match not in detected:
          true_positives.append(1)
          false_positives.append(0)
          detected.append(gt_match)
          TP += 1
        else:
          true_positives.append(0)
          false_positives.append(1)

      FN = len(current_labels) - TP
      false_negatives.append(FN)

    # Convert to cumulative sums
    tp_cumsum = np.cumsum(true_positives)
    fp_cumsum = np.cumsum(false_positives)

    # Compute recall and precision
    recall = tp_cumsum / count_labels
    precision = tp_cumsum / (tp_cumsum + fp_cumsum)

    # Compute AP using the trapezoidal rule (integrating precision over recall)
    ap = np.trapz(precision, recall)
    return ap, np.sum(false_negatives) / count_labels

  def threshold(self, class_name, max_false_negative):
    list_ap = []
    list_fn = []
    list_thresholds = np.linspace(0.0, 1.0, 101).tolist()
    for i, iou_threshold in enumerate(list_thresholds):
      ap, FN = self.compute_ap(class_name, iou_threshold)
      #debug("AP", ap, "FN", FN)
      if FN > max_false_negative:
        ap = 0.0
      list_ap.append(ap)
      list_fn.append(FN)
    idx = list_ap.index(max(list_ap))
    debug("Threshold", class_name, list_thresholds[idx], "AP", list_ap[idx], "FN", list_fn[idx])
    return list_thresholds[idx], list_ap[idx], list_fn[idx]

  def evaluate(self, iou_thresholds={"default": 0.0}):
    """
    Calculate metrics like mAP based on IoU.
    """
    aps = []
    class_ap = {}
    for class_name in self.class_names:
      ap, false_negatives = self.compute_ap(class_name, iou_thresholds.get(class_name, iou_thresholds["default"]))
      aps.append(ap)
      class_ap[class_name] = ap

    mAP = np.mean(aps)
    return {'mAP': mAP, 'AP per class': class_ap}
